/*
	graph.c
*/

#include <gdk/gdk.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "graph.h"
#include "interface.h"
#include "main.h"


GdkWindow *win;
GtkWidget *viewport;
GdkGC *gc;
GdkColor black, white, red = { 0, 65535, 0, 0 };
GdkFont *bold_font, *normal_font;
graph_t *fileop;

void graph_init (GtkWidget *widget)
{
	GdkColormap *cmap;
	GtkWidget *drawingarea;

	drawingarea = GTK_WIDGET (gtk_object_get_data
				(GTK_OBJECT (widget), "drawingarea"));
	viewport = GTK_WIDGET (gtk_object_get_data
				(GTK_OBJECT (widget), "viewport"));
	win = drawingarea->window;
	cmap = gdk_colormap_get_system ();

	gc = gdk_gc_new (win);
	gdk_color_black (cmap, &black);
	gdk_color_white (cmap, &white);
	if (!gdk_colormap_alloc_color (cmap, &red, FALSE, TRUE))
		g_error ("gt: couldn't allocate red\n");

	gdk_gc_set_background (gc, &white);
	bold_font = gdk_font_load (
			"-*-helvetica-bold-r-normal--12-*-*-*-p-*-*-*");
	normal_font = gdk_font_load (
			"-*-helvetica-medium-r-normal--12-*-*-*-p-*-*-*");
}

void graph_close (void)
{
	gdk_gc_unref (gc);
	gc = NULL;
}

graph_t *graph_create (void)
{
	graph_t *g;

	g = (graph_t *) g_malloc (sizeof (graph_t));
	g->vertices = NULL;
	g->edges = NULL;

	return g;
}

void graph_destroy (graph_t *g)
{
	graph_clear (g);
	g_free (g);
}

void graph_load (graph_t *g)
{
	GtkWidget *fileselector;

	fileop = g;

	fileselector = gtk_file_selection_new ("Open File");
	gtk_file_selection_complete (GTK_FILE_SELECTION (fileselector),
							"*.graph");
	gtk_window_set_modal (GTK_WINDOW (fileselector), TRUE);
	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION
			(fileselector)->ok_button), "clicked",
			GTK_SIGNAL_FUNC (graph_load_file), fileselector);
	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION
			(fileselector)->ok_button), "clicked",
			GTK_SIGNAL_FUNC (gtk_widget_destroy),
			GTK_OBJECT (fileselector));
	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION
			(fileselector)->cancel_button), "clicked",
			GTK_SIGNAL_FUNC (gtk_widget_destroy),
			GTK_OBJECT (fileselector));

	gtk_widget_show (fileselector);
}

void graph_load_file (GtkWidget *button, gpointer data)
{
	gchar *fname, *buf, *c;
	GtkFileSelection *fs = GTK_FILE_SELECTION (data);
	FILE *f;
	int num, i;
	vertex_t *v;

	fname = gtk_file_selection_get_filename (fs);

	f = fopen (fname, "r");
	if (!f) {
		modal_message ("Uh, oh!", "Loading failed!");
		return;
	}

	graph_clear (fileop);
	buf = (gchar *) g_malloc (64);

	/* FIXME: Not particularly safe! */

	/* Read in vertices */
	fscanf (f, "v=%i\n", &num);
	while (num--) {
		v = (vertex_t *) g_malloc (sizeof (vertex_t));
		fgetc (f);
		for (c=buf-1; *c != '"';)
			*++c = fgetc (f);
		*c = '\0';
		v->label = g_strdup (buf);
		fscanf (f, " %i %i\n", &v->x, &v->y);
		graph_add_vertex (fileop, v);
	}

	/* Read in edges */
	fscanf (f, "e=%i\n", &num);
	while (num--) {
		edge_t *e = (edge_t *) g_malloc (sizeof (edge_t));
		fgetc (f);
		for (c=buf-1; *c != '"';)
			*++c = fgetc (f);
		*c = '\0';

		v = graph_find_vertex_by_name (fileop, buf);
		if (!v) {
			modal_message ("Uh, oh!", "Couldn't find a vertex"
					" for an edge. Ignoring edge.");
			g_free (e);
			continue;
		}
		e->from = v;

		while (fgetc (f) != '"')
			;
		for (c=buf-1; *c != '"';)
			*++c = fgetc (f);
		*c = '\0';

		v = graph_find_vertex_by_name (fileop, buf);
		if (!v) {
			modal_message ("Uh, oh!", "Couldn't find a vertex"
					" for an edge. Ignoring edge.");
			g_free (e);
			continue;
		}
		e->to = v;

		fscanf (f, " %i %i\n", &i, &e->weight);
		e->directed = i ? TRUE : FALSE;

		graph_add_edge (fileop, e);
	}

	g_free (buf);
	fclose (f);
	graph_draw (fileop);
}

void graph_save (graph_t *g)
{
	GtkWidget *fileselector;

	fileop = g;

	fileselector = gtk_file_selection_new ("Save File");
	gtk_file_selection_complete (GTK_FILE_SELECTION (fileselector),
							"*.graph");
	gtk_window_set_modal (GTK_WINDOW (fileselector), TRUE);
	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION
			(fileselector)->ok_button), "clicked",
			GTK_SIGNAL_FUNC (graph_save_file), fileselector);
	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION
			(fileselector)->ok_button), "clicked",
			GTK_SIGNAL_FUNC (gtk_widget_destroy),
			GTK_OBJECT (fileselector));
	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION
			(fileselector)->cancel_button), "clicked",
			GTK_SIGNAL_FUNC (gtk_widget_destroy),
			GTK_OBJECT (fileselector));

	gtk_widget_show (fileselector);
}

void graph_save_file (GtkWidget *button, gpointer data)
{
	gchar *fname;
	GtkFileSelection *fs = GTK_FILE_SELECTION (data);
	FILE *f;
	GSList *l;

	fname = gtk_file_selection_get_filename (fs);

	f = fopen (fname, "w");
	if (!f) {
		modal_message ("Uh, oh!", "Saving failed!");
		return;
	}

	/* Dump vertices */
	fprintf (f, "v=%i\n", g_slist_length (fileop->vertices));
	for (l = fileop->vertices; l; l = l->next) {
		vertex_t *v = (vertex_t *) l->data;
		fprintf (f, "\"%s\" %i %i\n", v->label, v->x, v->y);
	}

	/* Dump edges */
	fprintf (f, "e=%i\n", g_slist_length (fileop->edges));
	for (l = fileop->edges; l; l = l->next) {
		edge_t *e = (edge_t *) l->data;
		fprintf (f, "\"%s\" \"%s\" %i %i\n", e->from->label,
			e->to->label, e->directed ? 1 : 0, e->weight);
	}

	fclose (f);
}

void graph_clear (graph_t *g)
{
	g_slist_foreach (g->vertices, (GFunc) graph_v_remover, g);
	g_slist_free (g->vertices);
	g->vertices = NULL;
	g->edges = NULL;

	gtk_widget_queue_draw (mainWin);
}

void graph_draw (graph_t *g)
{
	GtkWidget *it;
	gint w, h;
	gboolean opt;

	gdk_gc_set_foreground (gc, &white);
	gdk_window_get_size (win, &w, &h);
	gdk_draw_rectangle (win, gc, TRUE, 0, 0, w, h);

	it = gtk_object_get_data (GTK_OBJECT (mainWin), "view-labels");
	opt = GTK_CHECK_MENU_ITEM (it)->active;

	gdk_gc_set_foreground (gc, &black);
	g_slist_foreach (g->vertices, (GFunc) graph_v_drawer,
							(gpointer) opt);

	it = gtk_object_get_data (GTK_OBJECT (mainWin), "view-weights");
	opt = GTK_CHECK_MENU_ITEM (it)->active;

	g_slist_foreach (g->edges, (GFunc) graph_e_drawer,
							(gpointer) opt);
}

void graph_add_vertex (graph_t *g, vertex_t *v)
{
	char buf[4];

	if (!v->label) {
		/* If no label given, allocate one */
		buf[0] = 'A';
		buf[1] = '\0';
		do {
			if (!graph_find_vertex_by_name (g, buf))
				break;
			buf[0]++;
		} while (buf[0] < 'Z');
		v->label = g_strdup (buf);
		graph_add_vertex (g, v);
		return;
	} else if (graph_find_vertex_by_name (g, v->label)) {
		/* Duplicate label */
		int len = strlen (v->label) + 5;
		gchar *lab = (gchar *) g_malloc (len);
		g_snprintf (lab, len, "%s-dup", v->label);
		g_free (v->label);
		v->label = lab;
		graph_add_vertex (g, v);
		return;
	}

	v->selected = FALSE;
	v->edges = NULL;
	v->degree = 0;
	g->vertices = g_slist_append (g->vertices, v);
}

void graph_add_edge (graph_t *g, edge_t *e)
{
	e->selected = FALSE;
	e->from->edges = g_slist_append (e->from->edges, e);
	e->from->degree++;
	e->to->edges = g_slist_append (e->to->edges, e);
	e->to->degree++;
	g->edges = g_slist_append (g->edges, e);
}

void graph_add_vertex_dup (graph_t *g, vertex_t *v)
{
	vertex_t *vnew;

	vnew = (vertex_t *) g_malloc (sizeof (vertex_t));
	vnew->label = g_strdup (v->label);
	vnew->x = v->x;
	vnew->y = v->y;
	vnew->mark = v->mark;
	graph_add_vertex (g, vnew);
}

void graph_add_edge_dup (graph_t *g, edge_t *e)
{
	edge_t *enew;

	enew = (edge_t *) g_malloc (sizeof (edge_t));
	enew->from = graph_find_vertex_by_name (g, e->from->label);
	enew->to = graph_find_vertex_by_name (g, e->to->label);
	enew->directed = e->directed;
	enew->weight = e->weight;
	enew->mark = e->mark;
	graph_add_edge (g, enew);
}

void graph_remove_vertex (graph_t *g, vertex_t *v)
{
	g->vertices = g_slist_remove (g->vertices, v);
	graph_v_remover (v, g);
}

void graph_remove_edge (graph_t *g, edge_t *e)
{
	e->from->edges = g_slist_remove (e->from->edges, e);
	e->from->degree--;
	e->to->edges = g_slist_remove (e->to->edges, e);
	e->to->degree--;
	g->edges = g_slist_remove (g->edges, e);
	graph_e_remover (e, NULL);
}

vertex_t *graph_find_vertex (graph_t *g, int x, int y)
{
	search_t search;
	GSList *res;

	search.type = SEARCH_POSITION;
	search.param1 = (void *) x;
	search.param2 = (void *) y;
	res = g_slist_find_custom (g->vertices, &search,
					(GCompareFunc) graph_v_finder);

	if (res)
		return (vertex_t *) res->data;
	return NULL;
}

vertex_t *graph_find_vertex_by_name (graph_t *g, char *label)
{
	GSList *l;
	vertex_t *v;

	for (l = g->vertices; l; l = l->next) {
		v = (vertex_t *) l->data;
		if (!strcmp (v->label, label))
			return v;
	}
	return NULL;
}

GSList *graph_find_selected (graph_t *g)
{
	search_t search;

	search.type = SEARCH_SELECTED;
	search.param1 = NULL;
	g_slist_find_custom (g->vertices, &search,
					(GCompareFunc) graph_v_finder);

	return (GSList *) search.param1;
}

/* ASSUME UNDIRECTED */
gboolean graph_is_adjacent (vertex_t *v1, vertex_t *v2)
{
	GSList *l;

	if (v1 == v2)
		return FALSE;		/* Assuming no loops */
	for (l = v1->edges; l; l = l->next) {
		edge_t *e = (edge_t *) l->data;
		if ((e->from == v2) || (e->to == v2))
			return TRUE;
	}
	return FALSE;
}

void graph_unselect_all (graph_t *g)
{
	GSList *l;

	for (l = g->vertices; l; l = l->next)
		((vertex_t *) l->data)->selected = FALSE;
	for (l = g->edges; l; l = l->next)
		((edge_t *) l->data)->selected = FALSE;
}

void graph_move_selected (graph_t *g, int dx, int dy)
{
	GSList *l;
	vertex_t *v;

	for (l = g->vertices; l; l = l->next) {
		v = (vertex_t *) l->data;
		if (!v->selected)
			continue;
		v->x += dx;
		v->y += dy;
	}
}


void graph_v_remover (gpointer data, gpointer user)
{
	vertex_t *v = (vertex_t *) data;
	graph_t *g = (graph_t *) user;
	GSList *l;

	/* remove adjacent edges */
	for (l = v->edges; l; l = v->edges)
		graph_remove_edge (g, (edge_t *) l->data);

	g_slist_free (v->edges);
	g_free (v->label);
	g_free (v);
}

void graph_e_remover (gpointer data, gpointer user)
{
	edge_t *e = (edge_t *) data;

	g_free (e);
}

void graph_v_drawer (gpointer data, gpointer user)
{
	vertex_t *v = (vertex_t *) data;
	gboolean labels = (gboolean) user;
	int rad = VERTEX_RADIUS;

	if (v->selected)
		gdk_gc_set_foreground (gc, &red);

	gdk_draw_arc ((GdkDrawable *) win, gc, TRUE, v->x - rad / 2,
				v->y - rad / 2, rad, rad, 0, 64 * 360);

	if (labels)
		gdk_draw_string ((GdkDrawable *) win, bold_font, gc,
					v->x + rad, v->y - rad, v->label);

	if (v->selected)
		gdk_gc_set_foreground (gc, &black);
}

void graph_e_drawer (gpointer data, gpointer user)
{
	edge_t *e = (edge_t *) data;
	gboolean weights = (gboolean) user;
	gchar buf[8];
	int rx, ry, dx, dy;
	double theta;
	GdkPoint tri[3];

	theta = atan2 (e->to->y - e->from->y, e->to->x - e->from->x);
	dx = floor (VERTEX_RADIUS * cos (theta) / 2);
	dy = floor (VERTEX_RADIUS * sin (theta) / 2);

	if (e->selected)
		gdk_gc_set_foreground (gc, &red);

	/* draw line */
	gdk_gc_set_line_attributes (gc, 1, GDK_LINE_SOLID, GDK_CAP_BUTT,
							GDK_JOIN_MITER);

	gdk_draw_line ((GdkDrawable *) win, gc, e->from->x + dx,
			e->from->y + dy, e->to->x - dx, e->to->y - dy);
	if (e->directed) {
		tri[0].x = e->to->x - dx;
		tri[0].y = e->to->y - dy;
		tri[1].x = e->to->x - dx - dx - dy;
		tri[1].y = e->to->y - dy + dx - dy;
		tri[2].x = e->to->x - dx - dx + dy;
		tri[2].y = e->to->y - dy - dx - dy;
		gdk_draw_polygon ((GdkDrawable *) win, gc, TRUE, tri, 3);
	}

	if (weights) {
		/* draw label */
		g_snprintf (buf, 8, "%i", e->weight);
		rx = (e->from->x + e->to->x) / 2 + dy;
		ry = (e->from->y + e->to->y) / 2 - dx;
		if (theta < 0)
			rx += gdk_string_width (normal_font, buf)
							* sin (theta);
		if (fabs (theta) > M_PI_2)
			ry -= 2 * dx;
		gdk_draw_string ((GdkDrawable *) win, normal_font, gc,
							rx, ry, buf);
	}

	if (e->selected)
		gdk_gc_set_foreground (gc, &black);
}

gint graph_v_finder (gpointer data, gpointer user)
{
	vertex_t *v = (vertex_t *) data;
	search_t *s = (search_t *) user;
	int x = (int) s->param1, y = (int) s->param2;

	if (s->type == SEARCH_POSITION) {
		if ((abs (v->x - x) + abs (v->y - y)) <= VERTEX_RADIUS)
			return 0;	/* bingo! */
		return 1;
	} else if (s->type == SEARCH_SELECTED) {
		if (v->selected)
			s->param1 = g_slist_append (s->param1, v);
		return 1;
	}

	return 0;
}



#define UHOH(x)			\
	{			\
		g_warning x;	\
		oops++;		\
	}
#define UHOH_CONT(x)		\
	{			\
		g_warning x;	\
		oops++;		\
		continue;	\
	}

/* Returns zero on sanity */
int graph_sanity (graph_t *g)
{
	int oops, degsum;
	GSList *l1, *l2;
	vertex_t *v;
	edge_t *e, *e2;

	oops = 0;

	/* Check edges */
	for (l1=g->edges; l1; l1 = l1->next) {
		e = (edge_t *) l1->data;
		if (!e->from)
			UHOH_CONT (("e->from is NULL."))
		if (!e->to)
			UHOH_CONT (("e->to is NULL."))
		if (e->from == e->to)
			UHOH_CONT (("Loop from \"%s\" to itself!",
							e->from->label))

		v = graph_find_vertex_by_name (g, e->from->label);
		if (!v)
			UHOH_CONT (("e->from (\"%s\") not in G.",
							e->from->label))
		else if (e->from != v)
			UHOH_CONT (("e->from (\"%s\") not same as in G.",
							e->from->label))

		v = graph_find_vertex_by_name (g, e->to->label);
		if (!v)
			UHOH (("e->to (\"%s\") not in G.", e->to->label))
		else if (e->to != v)
			UHOH (("e->to (\"%s\") not same as in G.",
							e->to->label))

		for (l2=e->from->edges; l2; l2 = l2->next) {
			e2 = (edge_t *) l2->data;
			if (e == e2)
				continue;
			if (((e->from == e2->from) && (e->to == e2->to)) ||
			    ((e->from == e2->to) && (e->to == e2->from)))
				UHOH (("(\"%s\",\"%s\") is duplicated.",
					e->from->label, e->to->label))
		}
	}

	/* Check that all vertices have correct adjacency lists */
	degsum = 0;
	for (l1=g->vertices; l1; l1 = l1->next) {
		v = (vertex_t *) l1->data;
		degsum += g_slist_length (v->edges);

		/* Check degree */
		if (v->degree != g_slist_length (v->edges))
			UHOH (("\"%s\"'s degree count wrong!", v->label))

		/* Check for extraneous edges */
		for (l2=v->edges; l2; l2 = l2->next) {
			e = (edge_t *) l2->data;
			if ((e->from != v) && (e->to != v))
				UHOH (("Non-incident edge (%s,%s) in "
					"\"%s\"'s adjacency list.",
					e->from->label, e->to->label,
							v->label))
		}

		/* TODO: Check for missing edges */
	}
	if (degsum % 2)
		UHOH (("Degree sum is odd (%i)!", degsum))


	return oops;
}
