/*
	interface.c
*/

#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include <gtk/gtk.h>
#include "callbacks.h"
#include "config.h"
#include "graph.h"
#include "interface.h"
#include "main.h"
#include "matrix.h"
#include "op.h"
#include "prefab.h"


GtkWidget *create_mainWin (void)
{
	GtkWidget *win, *vbox, *menubar, *scrolledwindow, *viewport,
					*drawingarea, *statusbar, *it;
	GtkAccelGroup *accel_group;
	GtkItemFactory *factory;
	static GtkItemFactoryEntry menu_items[] = {
		{ "/_File", NULL, NULL, 0, "<Branch>" },
		{ "/File/_Open", "<control>O", graph_load, 0, NULL },
		{ "/File/_Save", "<control>S", graph_save, 0, NULL },
		{ "/File/", NULL, NULL, 0, "<Separator>" },
		{ "/File/E_xit", "<control>X", gtk_main_quit, 0, NULL },
		{ "/_View", NULL, NULL, 0, "<Branch>" },
		{ "/View/_Labels", NULL, graph_draw, 0, "<ToggleItem>" },
		{ "/View/_Weights", NULL, graph_draw, 0, "<ToggleItem>" },
		{ "/_Graph", NULL, NULL, 0, "<Branch>" },
		{ "/Graph/_Clear", "<control>W", graph_clear, 0, NULL },
		{ "/Graph/_Find", NULL, NULL, 0, "<Branch>" },
		{ "/Graph/Find/_Shortest Path", NULL, op_shortest_path,
								0, NULL },
		{ "/Graph/_Properties", NULL, NULL, 0, "<Branch>" },
		{ "/Graph/Properties/_Connectivity", NULL, op_connected,
								0, NULL },
		{ "/Graph/Properties/_Eulericity", "<control>E",
						op_eulerian, 0, NULL },
		{ "/Graph/Properties/_Hamiltonicity", NULL,
						op_hamiltonian, 0, NULL },
		{ "/Graph/Properties/_Planarity", NULL, op_planar, 0, NULL },
		{ "/Graph/_Statistics", NULL, NULL, 0, "<Branch>" },
		{ "/Graph/Statistics/_Adjacency Matrix", "<control>A",
						op_adj_matrix, 0, NULL },
		{ "/Graph/Statistics/_Degree Sequence", NULL,
					op_degree_sequence, 0, NULL },
		{ "/Graph/Statistics/D_iameter", NULL, op_diameter, 0, NULL },
		{ "/Graph/Statistics/_Girth", NULL, op_girth, 0, NULL },
		{ "/Graph/Statistics/", NULL, NULL, 0, "<Separator>" },
		{ "/Graph/Statistics/_Chromatic Number", NULL,
					op_chromatic_number, 0, NULL },
		{ "/Graph/Statistics/Chromatic _Polynomial", NULL,
					op_chromatic_poly, 0, NULL },
		{ "/Graph/Check Sanity", "<control>Z", check_sanity, 0, NULL },
		{ "/_Prefab", NULL, NULL, 0, "<Branch>" },
		{ "/Prefab/Complete (_Kn)", NULL, prefab_complete, 0,
								NULL },
		{ "/Prefab/Cycle (_Cn)", NULL, prefab_cycle, 0, NULL },
		{ "/Prefab/Null (_Nn)", NULL, prefab_null, 0, NULL },
		{ "/Prefab/Star (_Sn)", NULL, prefab_star, 0, NULL },
		{ "/Prefab/Tree (_Tn)", NULL, NULL, 0, NULL },
		{ "/Prefab/Wheel (_Wn)", NULL, prefab_wheel, 0, NULL },
		{ "/_Help", NULL, NULL, 0, "<LastBranch>" },
		{ "/Help/_About", NULL, about_activate, 0, NULL },
	};
	gint num_items = sizeof (menu_items) / sizeof (menu_items[0]);

	/* Create main window */
	win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (win), "GraphThing " GT_VERSION);
	gtk_window_set_default_size (GTK_WINDOW (win), 500, 400);
	gtk_window_set_position (GTK_WINDOW (win), GTK_WIN_POS_CENTER);
	gtk_signal_connect (GTK_OBJECT (win), "delete_event",
				GTK_SIGNAL_FUNC (mainWin_delete), NULL);
	gtk_signal_connect (GTK_OBJECT (win), "destroy",
				GTK_SIGNAL_FUNC (mainWin_destroy), NULL);

	/* Fill it with a vertical container box */
	vbox = gtk_vbox_new (FALSE, 0);
	gtk_widget_ref (vbox);
	gtk_object_set_data_full (GTK_OBJECT (win), "vbox", vbox,
				(GtkDestroyNotify) gtk_widget_unref);
	gtk_container_add (GTK_CONTAINER (win), vbox);
	gtk_widget_show (vbox);

	/* Add a menubar -- note use of mainGraph!! */
	accel_group = gtk_accel_group_new ();
	factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>",
							accel_group);
	gtk_item_factory_create_items (factory, num_items, menu_items,
							mainGraph);
	gtk_window_add_accel_group (GTK_WINDOW (win), accel_group);
	menubar = gtk_item_factory_get_widget (factory, "<main>");
	gtk_widget_ref (menubar);
	gtk_object_set_data_full (GTK_OBJECT (win), "menubar", menubar,
				(GtkDestroyNotify) gtk_widget_unref);
	gtk_box_pack_start (GTK_BOX (vbox), menubar, FALSE, FALSE, 0);
	gtk_widget_show (menubar);

	it = gtk_item_factory_get_widget (factory, "/View/Labels");
	gtk_object_set_data (GTK_OBJECT (win), "view-labels", it);
	it = gtk_item_factory_get_widget (factory, "/View/Weights");
	gtk_object_set_data (GTK_OBJECT (win), "view-weights", it);

	/* Add a scrolled window */
	scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (
			GTK_SCROLLED_WINDOW (scrolledwindow),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_widget_ref (scrolledwindow);
	gtk_object_set_data_full (GTK_OBJECT (win), "scrolledwindow",
		scrolledwindow, (GtkDestroyNotify) gtk_widget_unref);
	gtk_box_pack_start (GTK_BOX (vbox), scrolledwindow, TRUE, TRUE, 0);
	gtk_widget_show (scrolledwindow);

	/* Fill it with a viewport */
	viewport = gtk_viewport_new (NULL, NULL);
	gtk_widget_ref (viewport);
	gtk_object_set_data_full (GTK_OBJECT (win), "viewport",
			viewport, (GtkDestroyNotify) gtk_widget_unref);
	gtk_container_add (GTK_CONTAINER (scrolledwindow), viewport);
	gtk_widget_show (viewport);

	/* Fill it (the viewport) with a drawing area */
	drawingarea = gtk_drawing_area_new ();
	gtk_widget_ref (drawingarea);
	gtk_object_set_data_full (GTK_OBJECT (win), "drawingarea",
			drawingarea, (GtkDestroyNotify) gtk_widget_unref);
	gtk_container_add (GTK_CONTAINER (viewport), drawingarea);
	gtk_widget_show (drawingarea);
	gtk_widget_set_events (GTK_WIDGET (drawingarea),
		GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK);
	gtk_signal_connect (GTK_OBJECT (drawingarea), "expose_event",
			GTK_SIGNAL_FUNC (drawingarea_expose), NULL);
	gtk_signal_connect (GTK_OBJECT (drawingarea), "button_press_event",
			GTK_SIGNAL_FUNC (drawingarea_button_press), NULL);
	gtk_signal_connect (GTK_OBJECT (drawingarea), "motion_notify_event",
			GTK_SIGNAL_FUNC (drawingarea_motion_notify), NULL);

	/* Add a status bar */
	statusbar = gtk_statusbar_new ();
	gtk_widget_ref (statusbar);
	gtk_object_set_data_full (GTK_OBJECT (win), "statusbar",
			statusbar, (GtkDestroyNotify) gtk_widget_unref);
	gtk_box_pack_start (GTK_BOX (vbox), statusbar, FALSE, FALSE, 0);
	gtk_widget_show (statusbar);

	return win;
}

void create_popupMenu (void)
{
	GtkWidget *menu, *it;
	GtkItemFactory *factory;
	static GtkItemFactoryEntry menu_items[] = {
		{ "/Add _Edge", NULL, add_edge_activate, 0, NULL },
		{ "/_Remove Edge", NULL, remove_edge_activate, 0, NULL }
	};
	gint num_items = sizeof (menu_items) / sizeof (menu_items[0]);
	GSList *vertices;

	factory = gtk_item_factory_new (GTK_TYPE_MENU, "<main>", NULL);
	gtk_item_factory_create_items (factory, num_items, menu_items, NULL);
	menu = gtk_item_factory_get_widget (factory, "<main>");

	/* FIXME: Ugh! How horribly inefficient! */
	vertices = graph_find_selected (mainGraph);
	if (g_slist_length (vertices) != 2) {
		it = gtk_item_factory_get_widget (factory, "/Add Edge");
		gtk_widget_set_sensitive (it, FALSE);
		it = gtk_item_factory_get_widget (factory, "/Remove Edge");
		gtk_widget_set_sensitive (it, FALSE);
	} else {
		vertex_t *v1 = (vertex_t *) vertices->data,
			*v2 = (vertex_t *) vertices->next->data;

		if (graph_is_adjacent (v1, v2)) {
			it = gtk_item_factory_get_widget (factory, "/Add Edge");
			gtk_widget_set_sensitive (it, FALSE);
		} else {
			it = gtk_item_factory_get_widget (factory, "/Remove Edge");
			gtk_widget_set_sensitive (it, FALSE);
		}
	}
	g_slist_free (vertices);

	/* destroy menu when closed */
	gtk_signal_connect (GTK_OBJECT (menu), "hide",
				GTK_SIGNAL_FUNC (gtk_object_unref), menu);
	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 2, 0);
}

void create_matrixWindow (matrix_t *m, const gchar *title)
{
	GtkWidget *dlg, *vbox, *hbox1, *hbox2, *scrlwin, *vp, *tbl,
			*label, *sp, *aarea, *button;
	GtkObject *sp_adj;
	int i, j;
	gchar buf[16];
	GSList *l;

	dlg = gtk_dialog_new ();
	gtk_window_set_title (GTK_WINDOW (dlg), title);
	gtk_window_set_position (GTK_WINDOW (dlg), GTK_WIN_POS_CENTER);
	gtk_window_set_modal (GTK_WINDOW (dlg), TRUE);
	gtk_window_set_default_size (GTK_WINDOW (dlg), 300, 200);
	gtk_window_set_policy (GTK_WINDOW (dlg), FALSE, TRUE, FALSE);

	vbox = GTK_DIALOG (dlg)->vbox;
	hbox1 = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox1, TRUE, TRUE, 0);

	/* Top left section (table) */
	scrlwin = gtk_scrolled_window_new (NULL, NULL);
	gtk_box_pack_start (GTK_BOX (hbox1), scrlwin, TRUE, TRUE, 0);
	vp = gtk_viewport_new (NULL, NULL);
	gtk_container_add (GTK_CONTAINER (scrlwin), vp);

	tbl = gtk_table_new (m->order + 1, m->order + 1, FALSE);
	for (i=0, l=mainGraph->vertices; (i < m->order) && l;
						i++, l = l->next) {
		label = gtk_button_new_with_label
					(((vertex_t *) l->data)->label);
		gtk_table_attach (GTK_TABLE (tbl), label, i+1, i+2, 0, 1,
			(GTK_EXPAND | GTK_FILL), (GTK_EXPAND | GTK_FILL),
								0, 0);
		label = gtk_button_new_with_label
					(((vertex_t *) l->data)->label);
		gtk_table_attach (GTK_TABLE (tbl), label, 0, 1, i+1, i+2,
			(GTK_EXPAND | GTK_FILL), (GTK_EXPAND | GTK_FILL),
								0, 0);
	}
	for (i=0; i < m->order; i++)
	for (j=0; j < m->order; j++) {
		g_snprintf (buf, 16, "%i", MAT (m, i, j));
		label = gtk_label_new (g_strdup (buf));
		gtk_table_attach (GTK_TABLE (tbl), label, i+1, i+2, j+1, j+2,
			(GTK_EXPAND | GTK_FILL), (GTK_EXPAND | GTK_FILL),
								0, 0);
	}
	gtk_container_add (GTK_CONTAINER (vp), tbl);

	/* Top right section (exponent spin button) */
	hbox2 = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox1), hbox2, FALSE, FALSE, 0);
	label = gtk_label_new ("Exponent:");
	gtk_box_pack_start (GTK_BOX (hbox2), label, FALSE, FALSE, 10);
	sp_adj = gtk_adjustment_new (1, 1, 50, 1, 5, 5);
	sp = gtk_spin_button_new (GTK_ADJUSTMENT (sp_adj), 1, 0);
	gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (sp),
						GTK_UPDATE_IF_VALID);
#if 1
	gtk_widget_set_sensitive (label, FALSE);
	gtk_widget_set_sensitive (sp, FALSE);
#endif
	gtk_box_pack_start (GTK_BOX (hbox2), sp, TRUE, TRUE, 0);

	/* Bottom section (close button) */
	aarea = GTK_DIALOG (dlg)->action_area;
	button = gtk_button_new_with_label ("Close");
	gtk_box_pack_start (GTK_BOX (aarea), button, TRUE, FALSE, 0);
	GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default (button);
	gtk_widget_grab_focus (button);

	/* TODO: Update matrix display when exponent value is changed */

	gtk_signal_connect_object (GTK_OBJECT (dlg), "destroy",
			GTK_SIGNAL_FUNC (matrix_destroy), (gpointer) m);
	gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
		GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dlg));

	gtk_widget_show_all (dlg);
}

/* A simple modal message dialog with an "OK" button */
void modal_message (const gchar *title, const gchar *msg)
{
	GtkWidget *dialog, *label, *button;

	dialog = gtk_dialog_new ();
	gtk_window_set_title (GTK_WINDOW (dialog), title);
	gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER);
	gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
	gtk_window_set_policy (GTK_WINDOW (dialog), FALSE, FALSE, TRUE);

	label = gtk_label_new (msg);
	gtk_misc_set_padding (GTK_MISC (label), 30, 30);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), label,
						TRUE, TRUE, 0);
	gtk_widget_show (label);

	button = gtk_button_new_with_label ("OK");
	gtk_widget_set_usize (button, 80, -1);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
					button, FALSE, FALSE, 15);
	GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default (button);
	gtk_widget_grab_focus (button);
	gtk_widget_show (button);

	gtk_signal_connect_object (GTK_OBJECT (dialog), "destroy",
				GTK_SIGNAL_FUNC (gtk_main_quit), 0);
	gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
				GTK_SIGNAL_FUNC (gtk_widget_destroy),
						GTK_OBJECT (dialog));

	gtk_widget_show (dialog);
	gtk_main ();
}

void status_push (const gchar *msg, guint context)
{
	GtkStatusbar *statusbar;

	statusbar = GTK_STATUSBAR (gtk_object_get_data
				(GTK_OBJECT (mainWin), "statusbar"));
	gtk_statusbar_push (statusbar, context, msg);
}

gint status_pop (guint context)
{
	GtkStatusbar *statusbar;

	statusbar = GTK_STATUSBAR (gtk_object_get_data
				(GTK_OBJECT (mainWin), "statusbar"));
	gtk_statusbar_pop (statusbar, context);

	return FALSE;
}
