/*
	matrix.c
*/

#include <glib.h>
#include <math.h>
#include "math.h"
#include "matrix.h"


matrix_t *matrix_new (int order)
{
	matrix_t *m;

	m = (matrix_t *) g_malloc (sizeof (matrix_t));
	m->order = order;
	m->data = (int *) g_malloc0 (order * order * sizeof (int));
	return m;
}

void matrix_destroy (matrix_t *m)
{
	g_free (m->data);
	g_free (m);
}

void matrix_add (matrix_t *dst, matrix_t *a, matrix_t *b)
{
	int i, j, o = dst->order;

	for (i=0; i < o; i++)
	for (j=0; j < o; j++)
		MAT (dst, i, j) = MAT (a, i, j) + MAT (b, i, j);
}

void matrix_sub (matrix_t *dst, matrix_t *a, matrix_t *b)
{
	int i, j, o = dst->order;

	for (i=0; i < o; i++)
	for (j=0; j < o; j++)
		MAT (dst, i, j) = MAT (a, i, j) - MAT (b, i, j);
}

void matrix_mul (matrix_t *dst, matrix_t *a, matrix_t *b)
{
	int i, j, k, sum, o = dst->order;

	for (i=0; i < o; i++)
	for (j=0; j < o; j++) {
		sum = 0;
		for (k=0; k < o; k++)
			sum += MAT (a, i, k) * MAT (b, k, j);
		MAT (dst, i, j) = sum;
	}
}


poly_t *poly_new (int terms)
{
	poly_t *p;

	p = (poly_t *) g_malloc (sizeof (poly_t));
	p->terms = terms;
	p->data = (int *) g_malloc0 (sizeof (int) * terms);
	return p;
}

void poly_destroy (poly_t *p)
{
	g_free (p->data);
	g_free (p);
}

void poly_add (poly_t *dst, poly_t *a, poly_t *b)
{
	int i, cap;

	cap = MIN (a->terms, b->terms);
	for (i=0; i < cap; i++)
		dst->data[i] = a->data[i] + b->data[i];
	if (a->terms < b->terms) {
		for (i=cap; i < b->terms; i++)
			dst->data[i] = b->data[i];
	} else if (a->terms > b->terms) {
		for (i=cap; i < a->terms; i++)
			dst->data[i] = a->data[i];
	}
	for (; i < dst->terms; i++)
		dst->data[i] = 0;
}

void poly_sub (poly_t *dst, poly_t *a, poly_t *b)
{
	int i, cap;

	cap = MIN (a->terms, b->terms);
	for (i=0; i < cap; i++)
		dst->data[i] = a->data[i] - b->data[i];
	if (a->terms < b->terms) {
		for (; i < b->terms; i++)
			dst->data[i] = -(b->data[i]);
	} else if (a->terms > b->terms) {
		for (; i < a->terms; i++)
			dst->data[i] = a->data[i];
	}
	for (; i < dst->terms; i++)
		dst->data[i] = 0;
}

void poly_mul (poly_t *dst, poly_t *a, poly_t *b)
{
	int i, j;

	/* Operand scanning method */
	for (i=0; i < dst->terms; i++)
		dst->data[i] = 0;
	for (i=0; i < a->terms; i++)
		for (j=0; j < b->terms; j++)
			dst->data[i+j+1] += a->data[i] * b->data[j];
}

/* To get x*(x-1)^n, call "poly_binomial (foo, n, -1);" */
void poly_binomial (poly_t *dst, int power, int addend)
{
	int i, cap, run_add;

	cap = MIN (dst->terms, power + 1);
	run_add = addend;
	if (!(power & 1))
		run_add = -run_add;
	for (i=0; i < cap; i++) {
		dst->data[i] = run_add * math_choose (power, i);
		run_add *= addend;
	}
}

int poly_eval (poly_t *p, int x)
{
	int i, tot;

	tot = 0;
	for (i = (p->terms - 1); i >= 0; i--) {
		tot += p->data[i];
		tot *= x;
	}

	return tot;
}
