/*
	op.c
*/

#include <glib.h>
#include <gtk/gtk.h>
#include "graph.h"
#include "interface.h"
#include "main.h"
#include "matrix.h"
#include "op.h"


/* ASSUME UNDIRECTED */
void op_shortest_path (graph_t *g)
{
	GSList *l, *el;
	vertex_t *src, *dst, *min;
	int min_val, tot;
	gchar buf[32];

	/* Check if we have a source and a destination */
	l = graph_find_selected (g);
	if (g_slist_length (l) != 2) {
		modal_message ("Shortest Path", "Two vertices must be "
				"selected to find a shortest path.");
		return;
	}
	src = (vertex_t *) l->data;
	dst = (vertex_t *) l->next->data;

	status_push ("Finding shortest path...", 4);

	/* Method: Dijkstra's Algorithm
	 *	(vertex->mark is shortest distance from source + 1)
	 *	(if vertex->mark is negative, it's an estimate)
	 */

	/* Mark all vertices as belonging to the V\S set */
	for (l = g->vertices; l; l = l->next)
		((vertex_t *) l->data)->mark = (void *) 0;
	src->mark = (void *) 1;

	while (1) {
		/* Check whether every vertex is marked */
		for (l = g->vertices; l; l = l->next) {
			vertex_t *v = (vertex_t *) l->data;
			if (((int) v->mark) < 1)
				break;
		}
		if (!l)
			break;		/* all marked */

		/* Revise estimates */
		for (l = g->vertices; l; l = l->next) {
			vertex_t *v = (vertex_t *) l->data;
			int *us = (int *) &v->mark;
			if (*us > 0)
				continue;	/* already marked */
			*us = 0;
			for (el = v->edges; el; el = el->next) {
				edge_t *e = (edge_t *) el->data;
				vertex_t *alt = (e->from == v) ? e->to : e->from;
				int *them = (int *) &alt->mark;
				if (*them < 1)
					continue;
				if (((*them + e->weight) < -*us) || (*us == 0))
					*us = -(*them + e->weight);
			}
		}

		/* Find minimum estimate */
		min = src;
		min_val = 99999;	/* Ugh! */
		for (l = g->vertices; l; l = l->next) {
			vertex_t *v = (vertex_t *) l->data;
			int *us = (int *) &v->mark;
			if ((v == src) || (*us >= 0))
				continue;
			if (min_val > -*us) {
				min_val = -*us;
				min = v;
			}
		}
		if (min == src)
			break;		/* filled all possible */

		/* Mark this vertex */
		min->mark = (void *) -((int) min->mark);

		if (min == dst)
			break;		/* cut search short */
	}

	/* Read off data */
	tot = (int) dst->mark;

	/* Highlight shortest path */
	graph_unselect_all (g);
	dst->selected = TRUE;
	while (dst != src) {
		vertex_t *alt;
		edge_t *e;

		for (el = dst->edges; el; el = el->next) {
			e = (edge_t *) el->data;
			alt = (e->from == dst) ? e->to : e->from;
			if ((int) dst->mark == ((int) alt->mark + e->weight))
				break;
		}
		if (!el)
			break;		/* Uh, oh! */
		e->selected = TRUE;
		alt->selected = TRUE;
		dst = alt;
	}

	if (tot > 0) {
		g_snprintf (buf, 32, "Shortest path is %i.", tot - 1);
		modal_message ("Shortest Path", buf);
		graph_draw (g);
	} else
		modal_message ("Shortest Path", "No path found.");

	status_pop (4);
}

/* ASSUME UNDIRECTED */
void op_connected (graph_t *g)
{
	gboolean res;
	GSList *l;
	vertex_t *v;
	edge_t *e;

	res = op_is_connected (g);
	if (res) {
		modal_message ("Connectivity", "Graph is connected.");
		return;
	}

	/* Unconnected -- select marked vertices and edges */
	for (l = g->vertices; l; l = l->next) {
		v = (vertex_t *) l->data;
		v->selected = v->mark ? TRUE : FALSE;
	}
	for (l = g->edges; l; l = l->next) {
		e = (edge_t *) l->data;
		e->selected = e->mark ? TRUE : FALSE;
	}
	graph_draw (g);

	modal_message ("Connectivity", "Graph is not connected.");
}

/* ASSUME UNDIRECTED */
void op_eulerian (graph_t *g)
{
	GSList *l;
	int odd;

	for (l=g->vertices, odd=0; l; l = l->next)
		odd += ((vertex_t *) l->data)->degree & 1;

	if (odd == 0)
		modal_message ("Eulericity", "Graph is Eulerian.");
	else if (odd == 2)
		modal_message ("Eulericity", "Graph is semi-Eulerian.");
	else
		modal_message ("Eulericity",
			"Graph is not Eulerian,\nnor semi-Eulerian.");
}

/* ASSUME UNDIRECTED */
void op_hamiltonian (graph_t *g)
{
	int ham, n;

	ham = -1;	/* -1 is unknown, 0 is non-Ham, 1 is Ham */

	/* TODO: What about semi-Hamiltonian graphs? */

	if (!op_is_connected (g)) {
		ham = 0;
		goto ham_result;
	}

	n = g_slist_length (mainGraph->vertices);

	if (n >= 3) {
		/* Try Dirac's Theorem */
		GSList *l;
		for (l = g->vertices; l; l = l->next) {
			vertex_t *v = (vertex_t *) l->data;
			if ((2 * v->degree) < n)
				goto ham_dirac_exit;
		}

		/* Hamiltonian */
		ham = 1;
		goto ham_result;
	}
ham_dirac_exit:

	if (n >= 3) {
		/* Try Ore's Theorem */

		GSList *l1, *l2;
		for (l1 = g->vertices; l1; l1 = l1->next)
		for (l2 = l1->next; l2; l2 = l2->next) {
			vertex_t *v1 = (vertex_t *) l1->data,
				*v2 = (vertex_t *) l2->data;
			GSList *l3;
			for (l3 = v1->edges; l3; l3 = l3->next) {
				edge_t *e = (edge_t *) l3->data;
				if ((e->from == v2) || (e->to == v2))
					break;
			}
			if (l3)
				continue;	/* adjacent vertices */
			if ((v1->degree + v2->degree) < n)
				goto ham_ore_exit;
		}

		/* Hamiltonian */
		ham = 1;
		goto ham_result;
	}
ham_ore_exit:


ham_result:
	if (ham < 0)
		modal_message ("Hamiltonicity", "Unknown.");
	else if (ham == 0)
		modal_message ("Hamiltonicity",
					"Graph is non-Hamiltonian.");
	else
		modal_message ("Hamiltonicity", "Graph is Hamiltonian.");

	/* TODO: Highlight Hamiltonian cycle or path */
}

void op_planar (graph_t *g)
{
	modal_message ("WIP", "Not implemented.");
}

void op_adj_matrix (graph_t *g)
{
	matrix_t *m = op_get_adj_matrix (g);

	if (!m) {
		modal_message ("Adjacency Matrix",
				"Null graph has no adjacency matrix.");
		return;
	}

#if 0
	{
		int i, j;

		g_print ("{\n");
		for (i=0; i < m->order; i++) {
			g_print ("\t");
			for (j=0; j < m->order; j++)
				g_print ("%i ", MAT (m, i, j));
			g_print ("\n");
		}
		g_print ("}\n");
	}

	matrix_destroy (m);
#else
	create_matrixWindow (m, "Adjacency Matrix");

	/* matrix_destroy automatically called by dialog box */
#endif
}

/* ASSUME UNDIRECTED */
void op_degree_sequence (graph_t *g)
{
	GSList *deg, *l;
	GString *msg;
	int x;

	status_push ("Computing degree sequence...", 4);

	/* FIXME: Use a better collection than a GSList */

	deg = NULL;
	for (l = g->vertices; l; l = l->next) {
		x = ((vertex_t *) l->data)->degree;
		deg = g_slist_prepend (deg, (gpointer) x);
	}
	deg = g_slist_sort (deg, (GCompareFunc) op_compare);

	msg = g_string_new ("Degree Sequence:\n\n");
	for (l = deg; l; l = l->next) {
		char buf[8];
		if (l == deg)
			g_snprintf (buf, 8, "%i", (int) l->data);
		else
			g_snprintf (buf, 8, ", %i", (int) l->data);
		msg = g_string_append (msg, buf);
	}

	modal_message ("Degree Sequence", msg->str);

	g_string_free (msg, TRUE);
	g_slist_free (deg);

	status_pop (4);
}

void op_diameter (graph_t *g)
{
	modal_message ("WIP", "Not implemented.");
}

void op_girth (graph_t *g)
{
	modal_message ("WIP", "Not implemented.");
}

void op_chromatic_number (graph_t *g)
{
	poly_t *p;
	int i, max;
	char buf[16];

	status_push ("Computing chromatic number...", 4);

	p = op_calc_poly (g);
	max = g_slist_length (g->vertices);
	for (i=0; i < max; i++)
		if (poly_eval (p, i) > 0)
			break;

	poly_destroy (p);
	g_snprintf (buf, 16, "chi(G) = %i", i);
	modal_message ("Chromatic Number", buf);

	status_pop (4);
}

void op_chromatic_poly (graph_t *g)
{
	poly_t *poly;
	GString *msg;
	int i, cnt;

	status_push ("Computing chromatic polynomial...", 4);

	poly = op_calc_poly (g);

	msg = g_string_new ("Chromatic Polynomial:\n\n");
	cnt = 0;
	for (i=poly->terms; i > 0; i--) {
		char buf[16];
		int coeff = poly->data[i-1];

		if (coeff == 0)
			continue;
		if (cnt++ > 0) {
			if (coeff > 0)
				msg = g_string_append (msg, " + ");
			else {
				msg = g_string_append (msg, " - ");
				coeff = -coeff;
			}
		}
		if (i == 1)
			if (coeff == 1)
				g_snprintf (buf, 16, "x");
			else
				g_snprintf (buf, 16, "%ix", coeff);
		else {
			if (coeff == 1)
				g_snprintf (buf, 16, "x^%i", i);
			else
				g_snprintf (buf, 16, "%ix^%i", coeff, i);
		}
		msg = g_string_append (msg, buf);
	}
	if (poly->terms == 0)
		msg = g_string_append (msg, "0");

	modal_message ("Chromatic Polynomial", msg->str);

	g_string_free (msg, TRUE);
	poly_destroy (poly);

	status_pop (4);
}



gint op_compare (gpointer a, gpointer b)
{
	return ((int) a - (int) b);
}

/* ASSUME UNDIRECTED */
gboolean op_is_connected (graph_t *g)
{
	GSList *l;
	edge_t *e = NULL;
	gboolean done;

	if (!g->vertices)
		return TRUE;		/* null graph is connected */

	/* Unmark all vertices and edges initially */
	for (l = g->vertices; l; l = l->next)
		((vertex_t *) l->data)->mark = (void *) 0;
	for (l = g->edges; l; l = l->next)
		((edge_t *) l->data)->mark = (void *) 0;

	/* Mark a vertex */
	((vertex_t *) g->vertices->data)->mark = (void *) 1;

	/* Algorithm: Iterate through edge list, marking vertices adjacent
		to already-marked vertices, until no more can be marked
	 */
	do {
		done = TRUE;
		for (l = g->edges; l; l = l->next) {
			e = (edge_t *) l->data;
			if (e->mark)
				continue;
			if (!e->from->mark && !e->to->mark)
				continue;

			e->from->mark = e->to->mark = e->mark = (void *) 1;
			done = FALSE;
		}
	} while (!done);

	/* Check for any unmarked vertices */
	for (l = g->vertices; l; l = l->next)
		if (!((vertex_t *) l->data)->mark)
			return FALSE;
	return TRUE;
}

matrix_t *op_get_adj_matrix (graph_t *g)
{
	matrix_t *m;
	int v, i, j;
	GSList *l;

	v = g_slist_length (g->vertices);
	if (v < 1)
		return NULL;
	m = matrix_new (v);

	for (l = g->vertices, i=0; l; l = l->next, i++)
		((vertex_t *) l->data)->mark = (void *) i;

	for (l = g->edges; l; l = l->next) {
		edge_t *e = (edge_t *) l->data;
		i = (int) e->from->mark;
		j = (int) e->to->mark;
		MAT (m, i, j) = e->weight;
		if (!e->directed)
			MAT (m, j, i) = e->weight;
	}

	return m;
}

#if 0
void op_dump_graph (graph_t *g, char *name);
void op_dump_graph (graph_t *g, char *name)
{
	GSList *l;

	g_print ("%s: V={", name);
	for (l=g->vertices; l; l = l->next)
		g_print ("%s,", ((vertex_t *) l->data)->label);
	g_print ("\b} E={");
	for (l=g->edges; l; l = l->next) {
		edge_t *e = (edge_t *) l->data;
		g_print ("(%s,%s) ", e->from->label, e->to->label);
	}
	g_print ("\b}\n");
}

void op_dump_poly (poly_t *p);
void op_dump_poly (poly_t *p)
{
	int i;

	g_print ("\t");
	for (i = (p->terms - 1); i >= 0; i--)
		g_print ("%ix^%i ", p->data[i], i + 1);
	g_print ("\b\n");
}
#endif

/* ASSUME UNDIRECTED */
poly_t *op_calc_poly (graph_t *g)
{
	poly_t *res, *p1, *p2;
	int edge_num;
	graph_t *g1, *g2;
	GSList *l;
	vertex_t *v, *v2, *vnew;
	edge_t *e, *ealt, *enew;

	res = poly_new (g_slist_length (g->vertices));
	edge_num = g_slist_length (g->edges);

	/* Some quick, simple cases */
	if (res->terms == 0)
		return res;
	else if (res->terms == 1) {
		/* One vertex */
		res->data[0] = 1;	/* x */
		return res;
	} else if (res->terms == 2) {
		/* Two vertices */
		res->data[0] = (edge_num > 0) ? -1 : 0;
		res->data[1] = 1;	/* x^2 */
		return res;
	} else if (res->terms == 3) {
		/* Three vertices */
		res->data[2] = 1;		/* x^3 */
		res->data[1] = -edge_num;	/* x^2 */
		res->data[0] = (edge_num > 1) ? (edge_num - 1) : 0;	/* x */
		return res;
	} else if (edge_num < 1) {
		/* Null graphs */
		res->data[res->terms - 1] = 1;
		return res;
	}


	/* Firstly, if the graph is not connected, calculate polynomials
	 * for each component, then compute their product
	 */
	if (!op_is_connected (g)) {
		/* Split graph */
		g1 = graph_create ();		/* marked */
		g2 = graph_create ();		/* unmarked */
		for (l=g->vertices; l; l = l->next) {
			v = (vertex_t *) l->data;
			if (v->mark)
				graph_add_vertex_dup (g1, v);
			else
				graph_add_vertex_dup (g2, v);
		}
		for (l=g->edges; l; l = l->next) {
			e = (edge_t *) l->data;
			if (e->mark)
				graph_add_edge_dup (g1, e);
			else
				graph_add_edge_dup (g2, e);
		}

		/* Get chromatic polynomials of each component
		 * (yes, this works fine with more than 2 components)
		 */
		p1 = op_calc_poly (g1);
		p2 = op_calc_poly (g2);

		poly_mul (res, p1, p2);

		/* Clean up */
		poly_destroy (p1);
		poly_destroy (p2);
		graph_destroy (g1);
		graph_destroy (g2);

		return res;
	}

	/* We can now assume that we have a connected graph */

	/* Is it a tree? */
	if (edge_num == (res->terms - 1)) {
		/* poly = x(x-1)^(n-1) */
		poly_binomial (res, edge_num, -1);
		return res;
	}

	/* Special case for common 4-vertex connected graphs */
	if ((res->terms == 4) && (edge_num == 4)) {
		/* 4 vertices, 4 edges */
		e = (edge_t *) g->edges->data;
		ealt = (edge_t *) g->edges->next->data;
		res->data[3] = 1;
		res->data[2] = -4;
		if ((e->from->degree != e->to->degree) ||
		    (ealt->from->degree != ealt->to->degree)) {
			/* irregular -> x^4 - 4x^3 + 5x^2 - 2x */
			res->data[1] = 5;
			res->data[0] = -2;
		} else {
			/* regular (C4) -> x^4 - 4x^3 + 6x^2 - 3x */
			res->data[1] = 6;
			res->data[0] = -3;
		}
		return res;
	}
	if ((res->terms == 4) && (edge_num == 5)) {
		/* 4 vertices, 5 edges */
		/* must be (K4)-e -> x^4 - 5x^3 + 8x^2 - 4x */
		res->data[3] = 1;
		res->data[2] = -5;
		res->data[1] = 8;
		res->data[0] = -4;
		return res;
	}
	if ((res->terms == 4) && (edge_num == 6)) {
		/* K4 -> x^4 - 6x^3 + 11x^2 - 6x */
		res->data[3] = 1;
		res->data[2] = -6;
		res->data[1] = 11;
		res->data[0] = -6;
		return res;
	}

	/* Special case for K5 */
	if ((res->terms == 5) && (edge_num == 10)) {
		/* K5 -> x^5 - 10x^4 + 35x^3 - 50x^2 + 24x */
		res->data[4] = 1;
		res->data[3] = -10;
		res->data[2] = 35;
		res->data[1] = -50;
		res->data[0] = 24;
		return res;
	}

	/* TODO: *Really* handle complete graphs! */

	/* Use the chromatic reduction formula. We only use the
	 * edge-removal version, because it's easier to determine that a
	 * graph is a tree than if it's complete (yeah, I'm lazy...)
	 *		P(G) = P(G-e) - P(G|e)
	 */

	/* TODO: Also use the edge-addition version */

	/* TODO: Use a heavily connected edge, rather than just the first
	 *	(perhaps a bridge?)
	 */
	e = (edge_t *) g->edges->data;

	g1 = graph_create ();		/* remove edge (G-e) */
	g2 = graph_create ();		/* contract edge (G|e) */

	/* Populate remove-edge graph */
	for (l=g->vertices; l; l = l->next)
		graph_add_vertex_dup (g1, (vertex_t *) l->data);
	for (l=g->edges; l; l = l->next)
		if (e != (edge_t *) l->data)
			graph_add_edge_dup (g1, (edge_t *) l->data);

	/* Populate contract-edge graph */
	for (l=g->vertices; l; l = l->next)
		if ((l->data != e->from) && (l->data != e->to))
			graph_add_vertex_dup (g2, (vertex_t *) l->data);
	vnew = (vertex_t *) g_malloc (sizeof (vertex_t));
	vnew->label = NULL;
	graph_add_vertex (g2, vnew);
	for (l=g->edges; l; l = l->next) {
		ealt = (edge_t *) l->data;
		if (ealt == e)
			continue;
		if ((ealt->from != e->from) && (ealt->from != e->to) &&
		    (ealt->to != e->from) && (ealt->to != e->to)) {
			graph_add_edge_dup (g2, ealt);
			continue;
		}

		v = graph_find_vertex_by_name (g2, ealt->from->label);
		v2 = graph_find_vertex_by_name (g2, ealt->to->label);
		if (!v)
			v = vnew;
		if (!v2)
			v2 = vnew;
		if (graph_is_adjacent (v, v2))
			continue;

		enew = (edge_t *) g_malloc (sizeof (edge_t));
		enew->from = v;
		enew->to = v2;
		graph_add_edge (g2, enew);
	}

#if 0
	g_message ("Contracting (\"%s\",\"%s\") [new=\"%s\"].",
		e->from->label, e->to->label, vnew->label);
#endif
#if 0
	graph_sanity (g1);
	graph_sanity (g2);
#endif
	p1 = op_calc_poly (g1);
	p2 = op_calc_poly (g2);
	poly_sub (res, p1, p2);
#if 0
	op_dump_graph (g, "\tG");
	op_dump_poly (res);
	op_dump_graph (g1, "\tG-e");
	op_dump_poly (p1);
	op_dump_graph (g2, "\tG|e");
	op_dump_poly (p2);
#endif

	poly_destroy (p1);
	poly_destroy (p2);
	graph_destroy (g1);
	graph_destroy (g2);

	return res;
}
