/*
	prefab.c
*/

#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <math.h>
#include "graph.h"
#include "main.h"
#include "prefab.h"


void prefab_complete (graph_t *g)
{
	static int saved = 5;
	int param;
	GSList *l1, *l2;
	vertex_t *v1, *v2;
	edge_t *e;

	param = prefab_get_param ("Complete Graph (K_n):", 1, saved);
	if (param < 0)
		return;		/* Cancel */
	saved = param;

	prefab_null_do (g, param);
	if (param < 2)
		return;

	/* Add edges */
	for (l1=g->vertices; l1->next; l1 = l1->next) {
		v1 = (vertex_t *) l1->data;
		for (l2=l1->next; l2; l2 = l2->next) {
			v2 = (vertex_t *) l2->data;

			e = (edge_t *) g_malloc (sizeof (edge_t));
			e->from = v1;
			e->to = v2;
			e->directed = FALSE;
			e->weight = 1;
			graph_add_edge (g, e);
		}
	}

	graph_sanity (g);
}

void prefab_cycle (graph_t *g)
{
	static int saved = 5;
	int param;

	param = prefab_get_param ("Cycle Graph (C_n):", 1, saved);
	if (param < 0)
		return;		/* Cancel */
	saved = param;

	prefab_cycle_do (g, param);

	graph_sanity (g);
}

void prefab_null (graph_t *g)
{
	static int saved = 3;
	int param;

	param = prefab_get_param ("Null Graph (N_n):", 0, saved);
	if (param < 0)
		return;		/* Cancel */
	saved = param;

	prefab_null_do (g, param);
}

void prefab_star (graph_t *g)
{
	static int saved = 6;
	int param, w, h;
	GtkWidget *da;
	GSList *l;
	vertex_t *v, *vmid;
	edge_t *e;

	param = prefab_get_param ("Star Graph (S_n):", 1, saved);
	if (param < 0)
		return;		/* Cancel */
	saved = param;

	if (param <= 2) {
		prefab_cycle_do (g, param);
		return;
	}

	prefab_null_do (g, param - 1);

	da = GTK_WIDGET (gtk_object_get_data (GTK_OBJECT (mainWin),
							"drawingarea"));
	gdk_window_get_size (da->window, &w, &h);

	/* Add central vertex (hub) */
	vmid = (vertex_t *) g_malloc (sizeof (vertex_t));
	vmid->label = NULL;
	vmid->x = w / 2;
	vmid->y = h / 2;
	graph_add_vertex (g, vmid);

	/* Add edges */
	for (l=g->vertices; l->next; l = l->next) {
		v = (vertex_t *) l->data;

		e = (edge_t *) g_malloc (sizeof (edge_t));
		e->from = vmid;
		e->to = v;
		e->directed = FALSE;
		e->weight = 1;
		graph_add_edge (g, e);
	}

	graph_sanity (g);
}

void prefab_wheel (graph_t *g)
{
	static int saved = 6;
	int param, w, h;
	GtkWidget *da;
	GSList *l;
	vertex_t *v, *vmid;
	edge_t *e;

	param = prefab_get_param ("Wheel Graph (W_n):", 4, saved);
	if (param < 0)
		return;		/* Cancel */
	saved = param;

	prefab_cycle_do (g, param - 1);

	da = GTK_WIDGET (gtk_object_get_data (GTK_OBJECT (mainWin),
							"drawingarea"));
	gdk_window_get_size (da->window, &w, &h);

	/* Add central vertex (hub) */
	vmid = (vertex_t *) g_malloc (sizeof (vertex_t));
	vmid->label = NULL;
	vmid->x = w / 2;
	vmid->y = h / 2;
	graph_add_vertex (g, vmid);

	/* Add edges */
	for (l=g->vertices; l->next; l = l->next) {
		v = (vertex_t *) l->data;

		e = (edge_t *) g_malloc (sizeof (edge_t));
		e->from = vmid;
		e->to = v;
		e->directed = FALSE;
		e->weight = 1;
		graph_add_edge (g, e);
	}

	graph_sanity (g);
}



void prefab_cycle_do (graph_t *g, int param)
{
	edge_t *e;
	GSList *l;

	prefab_null_do (g, param);
	if (param < 2)
		return;

	/* Add edges */
	l = (param > 2) ? g->vertices : g->vertices->next;
	for (; l; l = l->next) {
		/* Connect to next vertex */
		e = (edge_t *) g_malloc (sizeof (edge_t));
		e->from = (vertex_t *) l->data;
		if (l->next)
			e->to = (vertex_t *) l->next->data;
		else
			e->to = (vertex_t *) g->vertices->data;
		e->directed = FALSE;
		e->weight = 1;
		graph_add_edge (g, e);
	}
}

void prefab_null_do (graph_t *g, int param)
{
	GtkWidget *da;
	int w, h, rad, i;
	vertex_t *v;

	da = GTK_WIDGET (gtk_object_get_data (GTK_OBJECT (mainWin),
							"drawingarea"));
	gdk_window_get_size (da->window, &w, &h);
	rad = MIN (w, h) * 3 / 8;	/* diameter = 75% */
	if (param == 1)
		rad = 0;		/* single vertex in middle */

	graph_clear (g);

	/* Add vertices in a circle, clockwise from top */
	for (i=0; i < param; i++) {
		v = (vertex_t *) g_malloc (sizeof (vertex_t));

		v->label = NULL;
		v->x = (w / 2) + rad * sin (2 * M_PI * i / param);
		v->y = (h / 2) - rad * cos (2 * M_PI * i / param);
		graph_add_vertex (g, v);
	}
}



int prefab_get_param (const gchar *text, int min, int initial)
{
	GtkWidget *dlg, *vbox, *hbox, *label, *sp, *aarea, *hbb,
						*buttonOK, *buttonCancel;
	GtkObject *sp_adj;
	GtkAccelGroup *accel_group;
	int ret = -1;

	initial = CLAMP (initial, min, 50);

	dlg = gtk_dialog_new ();
	gtk_object_set_data (GTK_OBJECT (dlg), "retval", &ret);
	gtk_window_set_title (GTK_WINDOW (dlg), "Prefab Parameter");
	gtk_window_set_position (GTK_WINDOW (dlg), GTK_WIN_POS_CENTER);
	gtk_window_set_modal (GTK_WINDOW (dlg), TRUE);
	gtk_window_set_default_size (GTK_WINDOW (dlg), 150, 150);
	gtk_window_set_policy (GTK_WINDOW (dlg), FALSE, TRUE, FALSE);

	vbox = GTK_DIALOG (dlg)->vbox;
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

	/* Top section (text and spinbutton) */
	label = gtk_label_new (text);
	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 10);
	sp_adj = gtk_adjustment_new (initial, min, 50, 1, 5, 5);
	sp = gtk_spin_button_new (GTK_ADJUSTMENT (sp_adj), 1, 0);
	gtk_object_set_data (GTK_OBJECT (dlg), "spinbutton", sp);
	gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (sp),
						GTK_UPDATE_IF_VALID);
	gtk_box_pack_start (GTK_BOX (hbox), sp, TRUE, TRUE, 0);

	/* Botton section (buttons) */
	aarea = GTK_DIALOG (dlg)->action_area;
	hbb = gtk_hbutton_box_new ();
	gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_SPREAD);
	gtk_box_pack_start (GTK_BOX (aarea), hbb, TRUE, TRUE, 0);
	buttonOK = gtk_button_new_with_label ("OK");
	GTK_WIDGET_SET_FLAGS (buttonOK, GTK_CAN_DEFAULT);
	gtk_container_add (GTK_CONTAINER (hbb), buttonOK);
	buttonCancel = gtk_button_new_with_label ("Cancel");
	gtk_container_add (GTK_CONTAINER (hbb), buttonCancel);

	accel_group = gtk_accel_group_new ();
	gtk_widget_add_accelerator (buttonOK, "clicked", accel_group,
						GDK_Return, 0, 0);
	gtk_widget_add_accelerator (buttonCancel, "clicked", accel_group,
						GDK_Escape, 0, 0);
	gtk_window_add_accel_group (GTK_WINDOW (dlg), accel_group);

	gtk_signal_connect_object (GTK_OBJECT (dlg), "destroy",
				GTK_SIGNAL_FUNC (gtk_main_quit), 0);
	gtk_signal_connect_object (GTK_OBJECT (buttonCancel), "clicked",
		GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dlg));
	gtk_signal_connect_object (GTK_OBJECT (buttonOK), "clicked",
		GTK_SIGNAL_FUNC (prefab_set_val), GTK_OBJECT (dlg));
	gtk_signal_connect_object (GTK_OBJECT (buttonOK), "clicked",
		GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT (dlg));

	gtk_widget_grab_default (buttonOK);
	gtk_widget_show_all (dlg);
	gtk_main ();

	return ret;
}

void prefab_set_val (GtkObject *obj)
{
	GtkSpinButton *sp;
	int *retval;

	sp = GTK_SPIN_BUTTON (gtk_object_get_data (obj, "spinbutton"));
	retval = (int *) gtk_object_get_data (obj, "retval");

	*retval = gtk_spin_button_get_value_as_int (sp);
}
